export abstract class Event {}

export type EventHandler<T> = (event: T) => void;

export type EventHandlerMap<TEventMap, T extends keyof TEventMap> = Record<
	T,
	EventHandler<TEventMap[T]>[]
>;

type Constructor = new (...args: any[]) => {};

export function EventEmitter<TEventMap, TBase extends Constructor>(
	Base: TBase
) {
	return class extends Base {
		private readonly listeners: EventHandlerMap<
			TEventMap,
			keyof TEventMap
		> = {} as EventHandlerMap<TEventMap, keyof TEventMap>;

		public on<T extends keyof TEventMap>(
			type: T,
			handler: EventHandler<TEventMap[T]>
		) {
			if (!this.listeners[type]) {
				this.listeners[type] = [];
			}
			this.listeners[type].push(handler);
		}

		public off<T extends keyof TEventMap>(
			type: T,
			handler: EventHandler<TEventMap[T]>
		) {
			if (this.listeners[type]) {
				this.listeners[type] = this.listeners[type].filter(
					(h) => h !== handler
				);
			}
		}

		public once<T extends keyof TEventMap>(
			type: T,
			handler: EventHandler<TEventMap[T]>
		) {
			const wrappedHandler = (event: TEventMap[T]) => {
				handler(event);
				this.off(type, wrappedHandler);
			};
			this.on(type, wrappedHandler);
		}

		public emit<T extends keyof TEventMap>(type: T, event: TEventMap[T]) {
			if (this.listeners[type]) {
				this.listeners[type].forEach((handler) => handler(event));
			}
		}
	};
}
